/* 
 * tclUnixFile.c --
 *
 *      This file contains wrappers around UNIX file handling functions.
 *      These wrappers mask differences between Windows and UNIX.
 *
 * Copyright (c) 1995 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * SCCS: @(#) tclUnixFile.c 1.48 97/07/07 16:38:11
 */

#include "tclInt.h"
#include "tclPort.h"

#include <unixlib.h>
#include <unixio.h>
#include <ctype.h>
#include <string.h>
#include <starlet.h>
#include <rms.h>
#if __VMS_VER >= 70000000
#include <pwd.h>
#endif


/*
 * The variable below caches the name of the current working directory
 * in order to avoid repeated calls to getcwd.  The string is malloc-ed.
 * NULL means the cache needs to be refreshed.
 */

static char *currentDir =  NULL;
static int currentDirExitHandlerSet = 0;

/*
 * The variable below is set if the exit routine for deleting the string
 * containing the executable name has been registered.
 */

static int executableNameExitHandlerSet = 0;

extern pid_t waitpid _ANSI_ARGS_((pid_t pid, int *stat_loc, int options));

/*
 * Static routines for this file:
 */

static void	FreeCurrentDir _ANSI_ARGS_((ClientData clientData));
static void	FreeExecutableName _ANSI_ARGS_((ClientData clientData));
static void	toLowerCase _ANSI_ARGS_((char *str));

/*
 *----------------------------------------------------------------------
 *
 * FreeCurrentDir --
 *
 *	Frees the string stored in the currentDir variable. This routine
 *	is registered as an exit handler and will be called during shutdown.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Frees the memory occuppied by the currentDir value.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static void
FreeCurrentDir(clientData)
    ClientData clientData;	/* Not used. */
{
    if (currentDir != (char *) NULL) {
        ckfree(currentDir);
        currentDir = (char *) NULL;
        currentDirExitHandlerSet = 0;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * FreeExecutableName --
 *
 *	Frees the string stored in the tclExecutableName variable. This
 *	routine is registered as an exit handler and will be called
 *	during shutdown.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Frees the memory occuppied by the tclExecutableName value.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static void
FreeExecutableName(clientData)
    ClientData clientData;	/* Not used. */
{
    if (tclExecutableName != (char *) NULL) {
        ckfree(tclExecutableName);
        tclExecutableName = (char *) NULL;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * TclChdir --
 *
 *	Change the current working directory.
 *
 * Results:
 *	The result is a standard Tcl result.  If an error occurs and 
 *	interp isn't NULL, an error message is left in interp->result.
 *
 * Side effects:
 *	The working directory for this application is changed.  Also
 *	the cache maintained used by TclGetCwd is deallocated and
 *	set to NULL.
 *
 *----------------------------------------------------------------------
 */

int
TclChdir(interp, dirName)
    Tcl_Interp *interp;		/* If non NULL, used for error reporting. */
    char *dirName;     		/* Path to new working directory. */
{
    char *e, *r, *u, uName[512];

    if (currentDir != NULL) {
	ckfree(currentDir);
	currentDir = NULL;
    }

    /*
     * Need to convert disk:[dir]file.dir to disk:[dir.file]
     *
     * Conver to Unix format and wing it from there
     */
    if (u = decc$translate_vms(dirName)) {
	/*
	 * We strip out the extension if it's there.
	 */
	strcpy(uName, u);
	u = uName;
	if (r = strrchr(u, '/'))
	    r++;
	else
	    r = u;
	if (e = strchr(r, '.'))
	    *e = 0;
    }
    else
	u = dirName;

    if (chdir(dirName) != 0) {
	if (interp != NULL) {
	    Tcl_AppendResult(interp, "couldn't change working directory to \"",
		    dirName, "\": ", Tcl_PosixError(interp), (char *) NULL);
	}
	return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * TclGetCwd --
 *
 *	Return the path name of the current working directory.
 *
 * Results:
 *	The result is the full path name of the current working
 *	directory, or NULL if an error occurred while figuring it out.
 *	The returned string is owned by the TclGetCwd routine and must
 *	not be freed by the caller.  If an error occurs and interp
 *	isn't NULL, an error message is left in interp->result.
 *
 * Side effects:
 *	The path name is cached to avoid having to recompute it
 *	on future calls;  if it is already cached, the cached
 *	value is returned.
 *
 *----------------------------------------------------------------------
 */

char *
TclGetCwd(interp)
    Tcl_Interp *interp;		/* If non NULL, used for error reporting. */
{
    char buffer[MAXPATHLEN+1];

    if (currentDir == NULL) {
        if (!currentDirExitHandlerSet) {
            currentDirExitHandlerSet = 1;
            Tcl_CreateExitHandler(FreeCurrentDir, (ClientData) NULL);
        }
#ifdef USEGETWD
	if ((int)getwd(buffer) == (int)NULL) {
	    if (interp != NULL) {
		Tcl_AppendResult(interp,
			"error getting working directory name: ",
			buffer, (char *)NULL);
	    }
	    return NULL;
	}
#else
	if (getcwd(buffer, MAXPATHLEN+1, 0) == NULL) {
	    if (interp != NULL) {
		if (errno == ERANGE) {
		    Tcl_SetResult(interp,
			    "working directory name is too long",
		            TCL_STATIC);
		} else {
		    Tcl_AppendResult(interp,
			    "error getting working directory name: ",
			    Tcl_PosixError(interp), (char *) NULL);
		}
	    }
	    return NULL;
	}
#endif
	currentDir = (char *) ckalloc((unsigned) (strlen(buffer) + 1));
	strcpy(currentDir, buffer);
    }
    return currentDir;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_FindExecutable --
 *
 *	This procedure computes the absolute path name of the current
 *	application, given its argv[0] value.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The variable tclExecutableName gets filled in with the file
 *	name for the application, if we figured it out.  If we couldn't
 *	figure it out, Tcl_FindExecutable is set to NULL.
 *
 *----------------------------------------------------------------------
 */

void
Tcl_FindExecutable(argv0)
    char *argv0;		/* The value of the application's argv[0]. */
{
    char *name;

    if (tclExecutableName != NULL) {
	ckfree(tclExecutableName);
	tclExecutableName = NULL;
    }

    tclExecutableName = (char *) ckalloc((unsigned) (strlen(argv0) + 1));
    strcpy(tclExecutableName, argv0);

    if (!executableNameExitHandlerSet) {
        executableNameExitHandlerSet = 1;
        Tcl_CreateExitHandler(FreeExecutableName, (ClientData) NULL);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * TclGetUserHome --
 *
 *	This function takes the passed in user name and finds the
 *	corresponding home directory specified in the password file.
 *
 * Results:
 *	The result is a pointer to a static string containing
 *	the new name.  If there was an error in processing the
 *	user name then the return value is NULL.  Otherwise the
 *	result is stored in bufferPtr, and the caller must call
 *	Tcl_DStringFree(bufferPtr) to free the result.
 *
 * Side effects:
 *	Information may be left in bufferPtr.
 *
 *----------------------------------------------------------------------
 */

char *
TclGetUserHome(name, bufferPtr)
    char *name;			/* User name to use to find home directory. */
    Tcl_DString *bufferPtr;	/* May be used to hold result.  Must not hold
				 * anything at the time of the call, and need
				 * not even be initialized. */
{
#if __VMS_VER >= 70000000
    struct passwd *pwPtr;

    pwPtr = getpwnam(name);
    if (pwPtr == NULL) {
#ifndef __VMS
	endpwent();
#endif	/* __VMS */
	 return NULL;
    }
    Tcl_DStringInit(bufferPtr);
    Tcl_DStringAppend(bufferPtr, pwPtr->pw_dir, -1);
#ifndef __VMS
    endpwent();
#endif	/* __VMS */
    return bufferPtr->string;
#else
    return NULL;
#endif
}

/*
 *----------------------------------------------------------------------
 *
 * TclMatchFiles --
 *
 *	This routine is used by the globbing code to search a
 *	directory for all files which match a given pattern.
 *
 * Results: 
 *	If the tail argument is NULL, then the matching files are
 *	added to the interp->result.  Otherwise, TclDoGlob is called
 *	recursively for each matching subdirectory.  The return value
 *	is a standard Tcl result indicating whether an error occurred
 *	in globbing.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

int
TclMatchFiles(interp, separators, dirPtr, pattern, tail)
    Tcl_Interp *interp;		/* Interpreter to receive results. */
    char *separators;		/* Path separators to pass to TclDoGlob. */
    Tcl_DString *dirPtr;	/* Contains path to directory to search. */
    char *pattern;		/* Pattern to match against. */
    char *tail;			/* Pointer to end of pattern. */
{
    char savedChar = '\0';
    char *patternEnd = tail;
    char *fs, *s;
    char *dirName, vName[512];
    char mask0[512], mask[512];
    char resultantString[NAM$C_MAXRSS];
    char expandedString[NAM$C_MAXRSS];
    int status;
    int baseLength = Tcl_DStringLength(dirPtr);
    int result = TCL_OK;
    struct FAB fab = cc$rms_fab;
    struct NAM nam = cc$rms_nam;
    struct stat statBuf;

    if (dirPtr->string[0] == '\0') {
	dirName = "sys$disk:[]";
    }
    else {
    	if (Tcl_UnixToVms(dirPtr->string, vName)) {
    	    dirName = vName;
    	}
    	else {
	    Tcl_AppendResult(interp, "couldn't read directory \"",
		    dirPtr->string, "\": illegal filename", (char *) NULL);
	    return TCL_ERROR;
    	}
    }

    if (*tail == '\\') {
    	tail++;
    }
    if (*tail == '\0') {
    	tail = NULL;
    } else {
    	tail++;
    }
    savedChar = *patternEnd;
    *patternEnd = '\0';
    strcpy(mask0, pattern);
    if (strcmp(mask0, "*") == 0)
    	strcpy(mask0, "*.*");
    toLowerCase(mask0);
    strcpy(mask, dirName);
    if (!strpbrk(mask, ":]"))
    	strcat(mask, "/");
    strcat(mask, mask0);
    
    fab.fab$l_fop = FAB$M_NAM;
    fab.fab$l_nam = &nam;
    fab.fab$l_fna = mask;
    fab.fab$b_fns = strlen(mask);

    nam.nam$b_rss = NAM$C_MAXRSS;
    nam.nam$l_rsa = resultantString;
    nam.nam$b_ess = NAM$C_MAXRSS;
    nam.nam$l_esa = expandedString;

    status = sys$parse(&fab);
    if (status != RMS$_NORMAL) {
    	if (baseLength > 0) {
    	    savedChar = dirPtr->string[baseLength-1];
    	    if (savedChar == '/') {
    	    	dirPtr->string[baseLength-1] = '\0';
	    }
	}
	errno = EVMSERR;
	vaxc$errno = status;
	Tcl_AppendResult(interp, "couldn't read directory \"",
		dirPtr->string, "\": ", Tcl_PosixError(interp), (char *) NULL);
	if (baseLength > 0) {
	    dirPtr->string[baseLength-1] = savedChar;
	}
	*patternEnd = savedChar;
	return TCL_ERROR;
    }

    /*
     * Clean up the end of the pattern and the tail pointer.  Leave
     * the tail pointing to the first character after the path separator
     * following the pattern, or NULL.  Also, ensure that the pattern
     * is null-terminated.
     */

    while ((status = sys$search(&fab)) == RMS$_NORMAL) {
	resultantString[nam.nam$b_rsl] = '\0';
	toLowerCase(resultantString);		/* Stop the shouting */

	/*
	 * Find end of the directory name
	 */
	if (fs = strchr(resultantString, ']'))
	    fs++;
	else if (fs = strchr(resultantString, ':'))
	    fs++;
	else
	    fs = resultantString;

	Tcl_DStringSetLength(dirPtr, baseLength);
	if (tail == NULL) {
	    Tcl_DStringAppend(dirPtr, fs, -1);
	    Tcl_AppendElement(interp, dirPtr->string);
	} else if ((stat(resultantString, &statBuf) == 0)
		&& S_ISDIR(statBuf.st_mode)) {
	    /*
	     * Modifying the resultant string kills the sys$search call
	     */
	    strcpy(mask0, fs);
	    if (s = strstr(mask0, ".dir;"))
		*s = 0;
	    Tcl_DStringAppend(dirPtr, mask0, -1);
	    Tcl_DStringAppend(dirPtr, "/", 1);
	    result = TclDoGlob(interp, separators, dirPtr, tail);
	    if (result != TCL_OK) {
		break;
	    }
	}
    }
    *patternEnd = savedChar;

    (void) sys$close(&fab);
    return result;
}

static void
toLowerCase(char *str)
{
    char c;
    while (c = *str) {
	if (isupper(c))
	    *str = tolower(c);
	str++;
    }
}

void 
Tcl_VmsTranslateFileName(Tcl_Interp *interp, char *name, Tcl_DString *bufferPtr)
{
    /*
     * Convert a filename to Unix format
     */
    char vmsName[L_tmpnam];
    char unixName[L_tmpnam];

    /*
     * Convert from Unix to VMS to normalize hybrid filenames.
     */
    if (strchr(name, '/')) {
    	if (Tcl_UnixToVms(name, vmsName)) {
    	}
    	else {
    	}
    }
    else
    	strcpy(vmsName, name);

    /*
     * Name is now is in VMS format, convert it back to Unix and return
     */
    Tcl_VmsToUnix(vmsName, unixName);
    
    Tcl_DStringInit(bufferPtr);
    Tcl_DStringAppend(bufferPtr, unixName, -1);
}	
