/*
 * tclVmsPort.h --
 *
 *	This header file handles porting issues that occur because
 *	of differences between systems.  It reads in UNIX-related
 *	header files and sets up UNIX-related macros for Tcl's UNIX
 *	core.  It should be the only file that contains #ifdefs to
 *	handle different flavors of UNIX.  This file sets up the
 *	union of all UNIX-related things needed by any of the Tcl
 *	core files.  This file depends on configuration #defines such
 *	as NO_DIRENT_H that are set up by the "configure" script.
 *
 *	Much of the material in this file was originally contributed
 *	by Karl Lehenbauer, Mark Diekhans and Peter da Silva.
 *
 *	This file assumes we're using Multinet on VMS.
 *
 * Copyright (c) 1991-1994 The Regents of the University of California.
 * Copyright (c) 1994-1995 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * SCCS: @(#) tclUnixPort.h 1.49 97/07/30 14:11:59
 */

#ifndef _TCLVMSPORT
#define _TCLVMSPORT

#ifdef USE_MULTINET
#define DONT_DECLARE_MALLOC
#endif	/* USE_MULTINET */

#ifndef _TCLINT
#   include "tclInt.h"
#endif
#include <errno.h>
#include <fcntl.h>
#ifdef HAVE_NET_ERRNO_H
#   include <net/errno.h>
#endif

#include <float.h>
#include <signal.h>
#include <stat.h>
#include <time.h>
#define unlink delete
#include <unistd.h>
#include <types.h>

/*
 * Socket support stuff: This likely needs more work to parameterize for
 * each system.
 */

#if defined(USE_MULTINET)
#include "multinet_root:[multinet.include.sys]types.h"
#include "multinet_root:[multinet.include.sys]socket.h"
#include "multinet_root:[multinet.include.netinet]in.h"
#include "multinet_root:[multinet.include.arpa]inet.h"
#include "multinet_root:[multinet.include]netdb.h"	/* gethostbyname() */
#elif defined(USE_UCX)
#include <socket.h>		/* struct sockaddr, SOCK_STREAM, ... */
#include <in.h>			/* struct in_addr, struct sockaddr_in */
#include <inet.h>		/* inet_ntoa() */
#include <netdb.h>		/* gethostbyname() */
#include <ioctl.h>
#define multinet_recv recv
#define multinet_send send
#define multinet_getpeername getpeername
#define multinet_getsockname getsockname
#define multinet_gethostbyaddr gethostbyaddr
#define multinet_gethostbyname gethostbyname
#define multinet_socket socket
#define multinet_setsockopt setsockopt
#define multinet_bind bind
#define multinet_listen listen
#define multinet_connect connect
#define multinet_inet_addr inet_addr
#define multinet_accept accept
#define socket_ioctl ioctl
#define socket_errno errno
#define socket_close close
#define socket_perror perror
#define vms_errno_string() strerror(errno)
#else	/* USE_MULTINET */
#include <sys/socket.h>		/* struct sockaddr, SOCK_STREAM, ... */
#include <netinet/in.h>		/* struct in_addr, struct sockaddr_in */
#include <arpa/inet.h>		/* inet_ntoa() */
#include <netdb.h>		/* gethostbyname() */
#endif	/* USE_MULTINET */

/*
 * NeXT doesn't define O_NONBLOCK, so #define it here if necessary.
 */

#ifndef O_NONBLOCK
#   define O_NONBLOCK 0x80
#endif

/*
 * HPUX needs the flag O_NONBLOCK to get the right non-blocking I/O
 * semantics, while most other systems need O_NDELAY.  Define the
 * constant NBIO_FLAG to be one of these
 */

#ifdef HPUX
#  define NBIO_FLAG O_NONBLOCK
#else
#  define NBIO_FLAG O_NDELAY
#endif

/*
 * The following defines denote malloc and free as the system calls
 * used to allocate new memory.  These defines are only used in the
 * file tclCkalloc.c.
 */

#define	TclpAlloc(size)		malloc(size)
#define TclpFree(ptr)           free(ptr)
#define TclpRealloc(ptr, size)  realloc(ptr, size)

/*
 * The default platform eol translation on Unix is TCL_TRANSLATE_LF:
 */

#define	TCL_PLATFORM_TRANSLATION	TCL_TRANSLATE_LF

/*
 * Not all systems declare the errno variable in errno.h. so this
 * file does it explicitly.  The list of system error messages also
 * isn't generally declared in a header file anywhere.
 */

extern int errno;

/*
 * The type of the status returned by wait varies from UNIX system
 * to UNIX system.  The macro below defines it:
 */

#ifdef _AIX
#   define WAIT_STATUS_TYPE pid_t
#else
#ifndef NO_UNION_WAIT
#   define WAIT_STATUS_TYPE union wait
#else
#   define WAIT_STATUS_TYPE int
#endif
#endif

/*
 * Supply definitions for macros to query wait status, if not already
 * defined in header files above.
 */

#ifndef WIFEXITED
#   define WIFEXITED(stat)  (((*((int *) &(stat))) & 0xff) == 0)
#endif

#ifndef WEXITSTATUS
#   define WEXITSTATUS(stat) (((*((int *) &(stat))) >> 8) & 0xff)
#endif

#ifndef WIFSIGNALED
#   define WIFSIGNALED(stat) (((*((int *) &(stat)))) && ((*((int *) &(stat))) == ((*((int *) &(stat))) & 0x00ff)))
#endif

#ifndef WTERMSIG
#   define WTERMSIG(stat)    ((*((int *) &(stat))) & 0x7f)
#endif

#ifndef WIFSTOPPED
#   define WIFSTOPPED(stat)  (((*((int *) &(stat))) & 0xff) == 0177)
#endif

#ifndef WSTOPSIG
#   define WSTOPSIG(stat)    (((*((int *) &(stat))) >> 8) & 0xff)
#endif

/*
 * Define constants for waitpid() system call if they aren't defined
 * by a system header file.
 */

#ifndef WNOHANG
#   define WNOHANG 1
#endif
#ifndef WUNTRACED
#   define WUNTRACED 2
#endif

/*
 * Supply macros for seek offsets, if they're not already provided by
 * an include file.
 */

#ifndef SEEK_SET
#   define SEEK_SET 0
#endif

#ifndef SEEK_CUR
#   define SEEK_CUR 1
#endif

#ifndef SEEK_END
#   define SEEK_END 2
#endif

#ifdef GETTOD_NOT_DECLARED
EXTERN int		gettimeofday _ANSI_ARGS_((struct timeval *tp,
			    struct timezone *tzp));
#endif

/*
 * Define access mode constants if they aren't already defined.
 */

#ifndef F_OK
#    define F_OK 00
#endif
#ifndef X_OK
#    define X_OK 01
#endif
#ifndef W_OK
#    define W_OK 02
#endif
#ifndef R_OK
#    define R_OK 04
#endif

/*
 * Define FD_CLOEEXEC (the close-on-exec flag bit) if it isn't
 * already defined.
 */

#ifndef FD_CLOEXEC
#   define FD_CLOEXEC 1
#endif

/*
 * On systems without symbolic links (i.e. S_IFLNK isn't defined)
 * define "lstat" to use "stat" instead.
 */

#if defined(VMS) && defined(S_IFLNK)
#undef S_IFLNK
#endif	/* VMS && S_IFLNK */
#if !defined(S_IFLNK) || defined(VMS)
#   define lstat stat
#ifdef S_ISLNK
#undef S_ISLNK
#endif
#endif

/*
 * Define macros to query file type bits, if they're not already
 * defined.
 */

#ifndef S_ISREG
#   ifdef S_IFREG
#       define S_ISREG(m) (((m) & S_IFMT) == S_IFREG)
#   else
#       define S_ISREG(m) 0
#   endif
# endif
#ifndef S_ISDIR
#   ifdef S_IFDIR
#       define S_ISDIR(m) (((m) & S_IFMT) == S_IFDIR)
#   else
#       define S_ISDIR(m) 0
#   endif
# endif
#ifndef S_ISCHR
#   ifdef S_IFCHR
#       define S_ISCHR(m) (((m) & S_IFMT) == S_IFCHR)
#   else
#       define S_ISCHR(m) 0
#   endif
# endif
#ifndef S_ISBLK
#   ifdef S_IFBLK
#       define S_ISBLK(m) (((m) & S_IFMT) == S_IFBLK)
#   else
#       define S_ISBLK(m) 0
#   endif
# endif
#ifndef S_ISFIFO
#   ifdef S_IFIFO
#       define S_ISFIFO(m) (((m) & S_IFMT) == S_IFIFO)
#   else
#       define S_ISFIFO(m) 0
#   endif
# endif
#ifndef S_ISLNK
#   ifdef S_IFLNK
#       define S_ISLNK(m) (((m) & S_IFMT) == S_IFLNK)
#   else
#       define S_ISLNK(m) 0
#   endif
# endif
#ifndef S_ISSOCK
#   ifdef S_IFSOCK
#       define S_ISSOCK(m) (((m) & S_IFMT) == S_IFSOCK)
#   else
#       define S_ISSOCK(m) 0
#   endif
# endif

/*
 * Make sure that MAXPATHLEN is defined.
 */

#ifndef MAXPATHLEN
#   ifdef PATH_MAX
#       define MAXPATHLEN PATH_MAX
#   else
#       define MAXPATHLEN 2048
#   endif
#endif

/*
 * Make sure that L_tmpnam is defined.
 */

#ifndef L_tmpnam
#   define L_tmpnam 100
#endif

/*
 * The following macro defines the type of the mask arguments to
 * select:
 */

#ifndef NO_FD_SET
#   define SELECT_MASK fd_set
#else
#   ifndef _AIX
	typedef long fd_mask;
#   endif
#   if defined(_IBMR2)
#	define SELECT_MASK void
#   else
#	define SELECT_MASK int
#   endif
#endif

/*
 * Define "NBBY" (number of bits per byte) if it's not already defined.
 */

#ifndef NBBY
#   define NBBY 8
#endif

/*
 * The following macro defines the number of fd_masks in an fd_set:
 */
#ifdef FD_SETSIZE
#undef FD_SETSIZE
#endif	/* FD_SETSIZE */
#define FD_SETSIZE 32

/*
    We can only select(wait for) from 32 sources(event flags)
*/
#define MASK_SIZE 1

/*
 * The following implements the Unix method for exiting the process.
 */
#define TclPlatformExit(status) exit(status)

/*
 * The following function always succeeds under Unix.
 */

#define TclHasSockets(interp) (TCL_OK)
#define	TclHasPipes() (0)

#define TclCreateCommandChannel(out, in, err, num, pidPtr)      NULL
#define TclClosePipeFile(x)

/*
 * Variables provided by the C library:
 */

#if defined(__sgi) || defined(__sgi)
#define environ _environ
#endif
extern char **environ;

/*
 * At present (12/91) not all stdlib.h implementations declare strtod.
 * The declaration below is here to ensure that it's declared, so that
 * the compiler won't take the default approach of assuming it returns
 * an int.  There's no ANSI prototype for it because there would end
 * up being too many conflicts with slightly-different prototypes.
 */

extern double strtod();

extern int getEventFlag();
extern void freeEventFlag(int);
extern int reserveEventFlag();

extern int TclVmsWatchSocket(void *, int);
extern int TclVmsWatchSemiSocket(void *, int);
extern int TclVmsWatchDevice(void *, int);
extern int VmsSocketReady(Tcl_Channel, int);
extern int VmsDeviceReady(Tcl_Channel, int);
extern Tcl_Channel Tcl_OpenDevChannel(Tcl_Interp *, char *, char *,
				      int, char *);
extern void Tcl_VmsTranslateFileName (Tcl_Interp *, char *, Tcl_DString *);

#ifdef USE_MULTINET
/*
 *  I keep getting the DECC prefixed names
 */
#define gethostbyname	multinet_gethostbyname
#define	getservbyname	multinet_getservbyname
#define getsockopt	multinet_getsockopt
#define setsockopt	multinet_setsockopt
extern char *multinet_inet_ntoa();
extern struct servent *multinet_getservbyname();
extern struct hostent *multinet_gethostbyname();
extern struct hostent *multinet_gethostbyaddr();
#endif	/* USE_MULTINET */

/*
 * The following macros define time related functions in terms of
 * standard Unix routines.
 */

#define TclpGetDate(t,u) localtime((t))
#define TclStrftime(s,m,f,t) (strftime((s),(m),(f),(t)))
#define TclpGetPid(pid)	    ((unsigned long) (pid))

#define	TclpReleaseFile(file)

extern int VMSstat _ANSI_ARGS_((char *path, struct stat *buf));
#ifdef stat
#undef stat
#endif
#define stat(path, bufPtr) VMSstat(path, bufPtr)
#ifdef lstat
#undef lstat
#endif
#define lstat(path, bufPtr) VMSstat(path, bufPtr)

extern int VMSaccess _ANSI_ARGS_((const char *filename, int mode));
#define access(path, mode) VMSaccess(path, mode)

extern int Tcl_UnixToVms(char *, char *);

/*
 * Returned by getDeviceTypeByName
 */
#define DEV_IS_SE	0	/* something else */
#define	DEV_IS_TERM	1	/* terminal */
#define	DEV_IS_MBX	2	/* mailbox */
#define DEV_IS_FILE	4	/* a file (maybe) */
#define DEV_IS_DISK	8	/* a disk */

#define DEV_WATCHING	16

extern int getDeviceType(int);
extern int getDeviceTypeByName(char *);

/*
 * What a file descriptor on VMS can be
 */
#define TCL_VMS_FILE		9
#define TCL_VMS_SOCKET		10
#define TCL_VMS_DEVICE		11
#define TCL_VMS_EVENTFLAG	12
#define TCL_VMS_SEMI_SOCKET	13

#endif /* _TCLVMSPORT */
