/* 
 * vmsdevchan.c
 *
 *	Channel driver for VMS channels based on devices
 *
 */

#include "tclInt.h"	/* Internal definitions for Tcl. */
#include "tclPort.h"	/* Portability features for Tcl. */
#include <unixio.h>
#include <unixlib.h>

#include <errno.h>
#include <ssdef.h>
#include <iodef.h>
#include <starlet.h>
#include <devdef.h>
#include <dvidef.h>
#include <msgdef.h>

typedef struct {
    int	l;
    void *a;
} DESC;

#define	LOAD_DESC(d, s) d.l=strlen(s);d.a=s

/*
 * The following structure is used to set or get the serial port
 * attributes in a platform-independant manner.
 */
 
typedef struct TtyAttrs {
    int baud;
    int parity;
    int data;
    int stop;
} TtyAttrs;

/*
 * This structure describes per-instance state of a device based channel.
 */

typedef struct DevState {
    Tcl_Channel channel;	/* Tcl channel */
    int flags;
    int chan;			/* VMS channel */
    int validMask;
    int eventFlag;		/* Set when mbox I/O completes */
    int checkMask;		/* what we're checking */
    int readyMask;		/* what's ready */
    short iosb[4];
    int mbxChan;		/* VMS mailbox channel for unsolicited input */
} DevState;

/*
 * Static routines for this file:
 */

static DevState *	CreateDev _ANSI_ARGS_((Tcl_Interp *interp,
			    char *name, char *mode, int perm));
static int		DevBlockModeProc _ANSI_ARGS_((
			    ClientData instanceData, int mode));
static int		DevCloseProc _ANSI_ARGS_((ClientData instanceData,
			    Tcl_Interp *interp));
static int		DevInputProc _ANSI_ARGS_((ClientData instanceData,
		            char *buf, int toRead, int *errorCode));
static int		DevOutputProc _ANSI_ARGS_((
			    ClientData instanceData, char *buf, int toWrite,
			    int *errorCode));
static void		DevWatchProc _ANSI_ARGS_((ClientData instanceData,
			    int mask));
static int		DevGetHandleProc _ANSI_ARGS_((ClientData instanceData,
			    int direction, ClientData *clientData));
static void		devFreeProc(ClientData);
static void		setVMSerrorCode(int);
static int		assignChan(char *, char *);
static int		getDeviceName(int, char *);
static int		crembx(char *);
static int		delmbx(int);
static int		cleanmbx(int);
static void		devReadyAstProc(ClientData);
static int		setWriteAttention(int, void (*)(ClientData), ClientData);

static int		TtyParseMode _ANSI_ARGS_((Tcl_Interp *interp,
			    CONST char *mode, int *speedPtr, int *parityPtr,
			    int *dataPtr, int *stopPtr));
static void		TtyGetAttributes _ANSI_ARGS_((int fd,
			    TtyAttrs *ttyPtr));
static int		TtyGetOptionProc _ANSI_ARGS_((ClientData instanceData, 
			    Tcl_Interp *interp, char *optionName,
			    Tcl_DString *dsPtr));
static void		TtyInit _ANSI_ARGS_((int fd));
static void		TtySetAttributes _ANSI_ARGS_((int fd,
			    TtyAttrs *ttyPtr));
static int		TtySetOptionProc _ANSI_ARGS_((ClientData instanceData,
			    Tcl_Interp *interp, char *optionName, 
			    char *value));

static void debugOut(char *);

/*
 * This structure describes the channel type structure for device based IO:
 */

static Tcl_ChannelType devChannelType = {
    "dev",				/* Type name. */
    DevBlockModeProc,			/* Set blocking/nonblocking mode.*/
    DevCloseProc,			/* Close proc. */
    DevInputProc,			/* Input proc. */
    DevOutputProc,			/* Output proc. */
    NULL,				/* Seek proc. */
    NULL,				/* Set option proc. */
    NULL,				/* Get option proc. */
    DevWatchProc,			/* Initialize notifier. */
    DevGetHandleProc			/* Get OS handles out of channel. */
};

/*
 * This structure describes the channel type structure for serial IO.
 * Note that this type is a subclass of the "file" type.
 */

static Tcl_ChannelType ttyChannelType = {
    "tty",				/* Type name. */
    DevBlockModeProc,			/* Set blocking/nonblocking mode.*/
    DevCloseProc,			/* Close proc. */
    DevInputProc,			/* Input proc. */
    DevOutputProc,			/* Output proc. */
    NULL,				/* Seek proc. */
    TtySetOptionProc,			/* Set option proc. */
    TtyGetOptionProc,			/* Get option proc. */
    DevWatchProc,			/* Initialize notifier. */
    DevGetHandleProc			/* Get OS handles out of channel. */
};

/*
 *----------------------------------------------------------------------
 *
 * DevBlockModeProc --
 *
 *	Set blocking or non-blocking mode on channel.
 *
 * Results:
 *	0 if successful, errno when failed.
 *
 * Side effects:
 *	Sets the device into blocking or non-blocking mode.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
DevBlockModeProc(instanceData, mode)
    ClientData instanceData;		/* Unused. */
    int mode;				/* The mode to set. */
{
    return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * DevInputProc --
 *
 *	Reads input from the IO channel into the buffer given. Returns
 *	count of how many bytes were actually read, and an error indication.
 *	This routine is used by the file and command pipeline drivers.
 *
 * Results:
 *	A count of how many bytes were read is returned and an error
 *	indication is returned in an output argument.
 *
 * Side effects:
 *	Reads input from the actual channel.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
DevInputProc(instanceData, buf, toRead, errorCode)
    ClientData instanceData;		/* Unused. */
    char *buf;				/* Where to store data read. */
    int toRead;				/* How much space is available
                                         * in the buffer? */
    int *errorCode;			/* Where to store error code. */
{
    short iosb[4];
    int chan, status;			/* The OS handle for reading. */
    int bytesRead;			/* How many bytes were actually
                                         * read from the input device? */
    DevState *statePtr = (DevState *)instanceData;

    *errorCode = 0;
    errno = 0;
    chan = statePtr->chan;
    
    /*
     * Assume there is always enough input available. This will block
     * appropriately, and read will unblock as soon as a short read is
     * possible, if the channel is in blocking mode. If the channel is
     * nonblocking, the read will never block.
     */
    
    if (toRead == 0) {
        return 0;
    }

    statePtr->readyMask &= ~TCL_READABLE;

    status = sys$qiow(0, chan, IO$_READVBLK, iosb, 0, 0,
		      buf, toRead, 0, 0, 0, 0);
    if ((status & 1) && ((status = iosb[0]) & 1)) {
	if (statePtr->flags & DEV_IS_TERM) {
	    if (iosb[1] == 0 && iosb[3] == 1 && iosb[2] == 26)
		return 0;
	    else
		iosb[1] += iosb[3];	/* Add in terminator length */
	}
	return iosb[1];
    }
    else {
	setVMSerrorCode(status);
	*errorCode = errno;
	return -1;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * DevOutputProc--
 *
 *	Writes the given output on the IO channel. Returns count of how
 *	many characters were actually written, and an error indication.
 *
 * Results:
 *	A count of how many characters were written is returned and an
 *	error indication is returned in an output argument.
 *
 * Side effects:
 *	Writes output on the actual channel.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static char	*newBuff = 0;
static int	newBuffSize = 4096*2;

static int
DevOutputProc(instanceData, buf, toWrite, errorCode)
    ClientData instanceData;		/* Unused. */
    char *buf;				/* The data buffer. */
    int toWrite;			/* How many bytes to write? */
    int *errorCode;			/* Where to store error code. */
{
    char *b;
    short iosb[4];
    int written = 0;
    int chan, status;
    int nWrite;
    DevState *statePtr = (DevState *) instanceData;

    *errorCode = 0;
    errno = 0;
    chan = statePtr->chan;

    /*
     * Terminals are special: need to add CR's to each LF output
     */
    if (statePtr->flags & DEV_IS_TERM) {
	char *in, *out, c;
	int i;

	if (newBuff == (char *)NULL)
	    newBuff = ckalloc(newBuffSize);

	if (toWrite > newBuffSize/2) {
	    newBuff = realloc(newBuff, toWrite*2);
	    newBuffSize = toWrite*2;
	}

	in = buf;
	b = out = newBuff;
	for (i=0, nWrite=0; i<toWrite; i++) {
	    c = *out++ = *in++;
	    nWrite++;
	    written++;

#if 0
	    if (nWrite >= sizeof(newBuff)) {
		status = sys$qiow(0, chan, IO$_WRITEVBLK, iosb, 0, 0,
				  b, nWrite, 0, 0, 0, 0);
		if ((status & 1) && ((status = iosb[0]) & 1)) {
		    nWrite = 0;
		    out = newBuff;
		}
		else {
		    setVMSerrorCode(status);
		    *errorCode = errno;
		    return -1;
		}
	    }
#endif
	    if (c == '\r') {
		*out++ = '\n';
		nWrite++;
	    }
	}
    }
    else {
	b = buf;
	nWrite = toWrite;
    }
    status = sys$qiow(0, chan, IO$_WRITEVBLK, iosb, 0, 0,
			  b, nWrite, 0, 0, 0, 0);
    if ((status & 1) && ((status = iosb[0]) & 1)) {
	if ((statePtr->flags & DEV_IS_TERM) && nWrite == iosb[1])
	    return written;
	else
	    return iosb[1];
    }
    else {
	setVMSerrorCode(status);
	*errorCode = errno;
	return -1;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * DevCloseProc --
 *
 *	Closes the IO channel.
 *
 * Results:
 *	0 if successful, the value of errno if failed.
 *
 * Side effects:
 *	Closes the physical channel
 *
 *----------------------------------------------------------------------
 */

static int
DevCloseProc(instanceData, interp)
    ClientData instanceData;	/* Unused. */
    Tcl_Interp *interp;
{
    int chan, errorCode = 0;
    DevState *statePtr = (DevState *) instanceData;

    if ((chan = statePtr->chan)) {
	Tcl_DeleteFileHandler(statePtr->eventFlag);
	if (sys$dassgn(chan) < 0) {
	    errorCode = errno;
	}
    }
    return errorCode;
}

/*
 *----------------------------------------------------------------------
 *
 * DevWatchProc --
 *
 *	Initialize the notifier to watch Tcl_Files from this channel.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Sets up the notifier so that a future event on the channel will
 *	be seen by Tcl.
 *
 *----------------------------------------------------------------------
 */

static void
DevWatchProc(instanceData, mask)
    ClientData instanceData;		/* The file state. */
    int mask;				/* Events of interest; an OR-ed
    					 * combination of TCL_READABLE,
    					 * TCL_WRITABLE and TCL_EXCEPTION. */
{
    DevState *statePtr = (DevState *) instanceData;

    mask &= statePtr->validMask;
    if (mask) {
	if (statePtr->eventFlag == 0)
	    statePtr->eventFlag = getEventFlag();
	sys$setast(0);
	statePtr->checkMask = mask;
	if (mask & TCL_READABLE) {
	    if (statePtr->readyMask & TCL_READABLE) {
		sys$setef(statePtr->eventFlag);
	    }
	    else {
		sys$clref(statePtr->eventFlag);
		statePtr->flags |= DEV_WATCHING;
	    }
	}
	sys$setast(1);
	Tcl_CreateFileHandler(statePtr->eventFlag, mask,
		TCL_VMS_DEVICE,
		(Tcl_FileProc *) Tcl_NotifyChannel,
		(ClientData) statePtr->channel);
    }
    else {
	Tcl_DeleteFileHandler(statePtr->eventFlag);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * DevGetHandleProc --
 *
 *	Called from Tcl_GetChannelFile to retrieve OS handles from
 *	a device-based channel.
 *
 * Results:
 *	Returns TCL_OK with the fd in handlePtr, or TCL_ERROR if
 *	there is no handle for the specified direction. 
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static int
DevGetHandleProc(instanceData, direction, handlePtr)
    ClientData instanceData;	/* The file state. */
    int direction;		/* TCL_READABLE or TCL_WRITABLE */
    ClientData *handlePtr;	/* Where to store the handle.  */
{
    DevState *fsPtr = (DevState *) instanceData;

    if (direction & fsPtr->validMask) {
	*handlePtr = (ClientData) fsPtr->chan;
	return TCL_OK;
    } else {
	return TCL_ERROR;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_OpenDevChannel --
 *
 *	Open an device based channel on VMS systems.
 *
 * Results:
 *	The new channel or NULL. If NULL, the output argument
 *	errorCodePtr is set to a POSIX error.
 *
 * Side effects:
 *	May open the channel
 *
 *----------------------------------------------------------------------
 */

Tcl_Channel
Tcl_OpenDevChannel(Tcl_Interp *interp, char *fileName, char *modeString, int permissions, char *chanName)
{
    char *s;
    char cName[20];
    char *channelName;
    int status;
    DevState *statePtr;
    int mode;
    int isAtty;

    statePtr = CreateDev(interp, fileName, modeString, permissions);
    if (statePtr == NULL)
	return NULL;

    if (chanName)
	channelName = chanName;
    else {
	sprintf(cName, "dev%d", statePtr->chan);
	channelName = cName;
    }
    
    mode = 0;
    if (strchr(modeString, 'r'))
    	mode |= TCL_READABLE;
    if (strchr(modeString, 'w'))
    	mode |= TCL_WRITABLE;

    if (statePtr->flags & DEV_IS_TERM) {
	statePtr->channel = Tcl_CreateChannel(&ttyChannelType, channelName,
		(ClientData) statePtr, mode);
	Tcl_SetChannelOption(interp, statePtr->channel, "-translation", "cr");
	Tcl_SetChannelOption(interp, statePtr->channel, "-buffering", "line");
    }
    else
	statePtr->channel = Tcl_CreateChannel(&devChannelType, channelName,
		(ClientData) statePtr, mode);
    return statePtr->channel;
}

static DevState *
CreateDev(Tcl_Interp *interp, char *name, char *modeString, int perm)
{
    char mbxName[65], *mbn;
    int status, mbxChan;
    int channel, seekFlag, mode, channelPermissions;
    DevState *statePtr;

    mode = TclGetOpenMode(interp, modeString, &seekFlag);
    if (mode == -1) {
        return NULL;
    }
    switch (mode & (O_RDONLY | O_WRONLY | O_RDWR)) {
	case O_RDONLY:
	    channelPermissions = TCL_READABLE;
	    break;
	case O_WRONLY:
	    channelPermissions = TCL_WRITABLE;
	    break;
	case O_RDWR:
	    channelPermissions = (TCL_READABLE | TCL_WRITABLE);
	    break;
	default:
	    Tcl_AppendResult(interp, "invalid mode value", (char *)NULL);
	    return NULL;
    }

    /*
     * 1. Create mailbox
     * 2. Get name of mailbox
     * 3. Create channel w/ mbox
     * 4. Set write attention on mbox
     */
    if (channelPermissions & TCL_READABLE) {
	mbxChan = crembx(mbxName);
	if (!mbxChan) {
	    if (interp) {
		Tcl_AppendResult(interp, "couldn't create a mailbox: ",
			Tcl_PosixError(interp), (char *) NULL);
	    }
	    return NULL;
	}
	mbn = mbxName;
    }
    else {
    	mbn = 0;
    	mbxChan = 0;
    }
    channel = assignChan(name, mbn);
    if (channel == 0) {
    	if (interp) {
	    Tcl_AppendResult(interp, "couldn't open \"", name, "\": ",
		    Tcl_PosixError(interp), (char *) NULL);
	}
	if (mbxChan)
	    delmbx(mbxChan);
	return NULL;
    }
    statePtr = (DevState *) ckalloc((unsigned) sizeof(DevState));
    statePtr->chan = channel;
    statePtr->flags = 0;
    statePtr->validMask = channelPermissions | TCL_EXCEPTION;
    statePtr->eventFlag = getEventFlag();
    statePtr->checkMask = 0;
    statePtr->readyMask = 0;
    statePtr->mbxChan = mbxChan;

    /*
     * Figure out what kind of device this is
     */
    status = getDeviceType(channel);
    if (status < 0) {
    	if (interp) {
	    Tcl_AppendResult(interp, "couldn't get device information:",
		    Tcl_PosixError(interp), (char *) NULL);
	}
	devFreeProc((ClientData)statePtr);
	return NULL;
    }
    else if (status == DEV_IS_TERM)
	statePtr->flags |= DEV_IS_TERM;
    else if (status == DEV_IS_MBX)
	statePtr->flags |= DEV_IS_TERM;
    else {
	/* who knows */
    }

    if (statePtr->flags & DEV_IS_TERM || statePtr->flags & DEV_IS_MBX) {
	/*
	 * Issue a "set write attention AST" to let us know
	 * when data is available to be read.
	 */
	if (channelPermissions & TCL_READABLE) {
	    if (!setWriteAttention(mbxChan, devReadyAstProc, statePtr)) {
	    	if (interp) {
		    Tcl_AppendResult(interp,
			"couldn't set write attention on device \"", name,
			"\":", Tcl_PosixError(interp), (char *) NULL);
		}
		devFreeProc((ClientData)statePtr);
		return 0;
	    }
	}
    }
    return statePtr;
}

static void
devReadyAstProc(ClientData clientdata)
{
    /*
     * AST handler invoked when there's data to be read
     */
    DevState *statePtr = (DevState *)clientdata;

    /*
     * Read the notification message that triggered the AST
     */
    cleanmbx(statePtr->mbxChan);

    if (statePtr->checkMask & TCL_READABLE) {
	statePtr->readyMask |= TCL_READABLE;
	if (statePtr->flags & DEV_WATCHING) {
	    sys$setef(statePtr->eventFlag);	/* For vmsselect */
	}
    }
    setWriteAttention(statePtr->mbxChan, devReadyAstProc, statePtr);
}

static void
devFreeProc(ClientData clientdata)
{
    /*
     * Free up device info structure
     */
    DevState *statePtr = (DevState *)clientdata;

    if (statePtr) {
	if (statePtr->eventFlag)
	    freeEventFlag(statePtr->eventFlag);
	if (statePtr->mbxChan)
	    delmbx(statePtr->mbxChan);
	ckfree((char *)statePtr);
    }
}

TclVmsWatchDevice(void *clientData, int mask)
{
    /*
     * Watch for something to happen on a device.
     */
    Tcl_Channel channel = (Tcl_Channel)clientData;
    DevState *statePtr = Tcl_GetChannelInstanceData(channel);

    sys$setast(0);
    if (statePtr->flags & (DEV_IS_TERM | DEV_IS_MBX)) {
	statePtr->checkMask = mask;
	if (mask & TCL_READABLE) {
	    if (statePtr->readyMask & TCL_READABLE) {
		sys$setef(statePtr->eventFlag);
	    }
	    else {
		sys$clref(statePtr->eventFlag);
		statePtr->flags |= DEV_WATCHING;
	    }
	    sys$setast(1);
	    return statePtr->eventFlag;
	}
    }
    sys$setast(1);
    return 0;
}

VmsDeviceReady(Tcl_Channel channel, int mask)
{
    /*
     * Check and see what's ready for a device
     */
    int readyMask = 0;
    DevState *statePtr = (DevState *)Tcl_GetChannelInstanceData(channel);

    sys$setast(0);
    if (statePtr->readyMask & TCL_READABLE)
	readyMask |= TCL_READABLE;
    if (statePtr->readyMask & TCL_WRITABLE)
	readyMask |= TCL_WRITABLE;
    if (statePtr->readyMask & TCL_EXCEPTION)
	readyMask |= TCL_EXCEPTION;
    statePtr->flags |= DEV_WATCHING;
    sys$setast(1);
    return readyMask;
}

int
Tcl_DevCmd(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    /*
     * TCL command for doing I/O on VMS devices.
     *
     * [dev deviceName ...]
     */
    char *modeString;
    int prot;
    Tcl_Channel chan;

    if (argc != 2 || argc > 4) {
	Tcl_AppendResult(interp, "Wrong # of arguments to \"dev\" command",
			 (char *)NULL);
	return TCL_ERROR;
    }
    prot = 0666;
    if (argc == 2) {
	modeString = "r";
    }
    else  {
	modeString = argv[2];
	if (argc == 4) {
	    if (Tcl_GetInt(interp, argv[3], &prot) != TCL_OK) {
		return TCL_ERROR;
	    }
	}
    }
    chan = Tcl_OpenDevChannel(interp, argv[1], modeString, prot, (char *)0);
    if (chan == (Tcl_Channel) NULL) {
	return TCL_ERROR;
    }
    Tcl_RegisterChannel(interp, chan);
    Tcl_AppendResult(interp, Tcl_GetChannelName(chan), (char *) NULL);
    return TCL_OK;
}

static void
setVMSerrorCode(int status)
{
    /*
     * Convert a VMS error number to a C error number
     */
#ifdef EVMSERR
    errno = EVMSERR;
#else
    errno = VMSERRNO;
#endif	/* EVMSERR */
    vaxc$errno = status;
}

static
assignChan(char *name, char *mbxName)
{
    short channel;
    int status;
    DESC d, mbxD;

    LOAD_DESC(d, name);
    if (mbxName) {
	LOAD_DESC(mbxD, mbxName);
	status = sys$assign(&d, &channel, 0, &mbxD, 0);
    }
    else
	status = sys$assign(&d, &channel, 0, 0, 0);
    if (status & 1)
	return channel;
    else {
    	printf("assignChan for %s failed: %x\n", name, status);
	setVMSerrorCode(status);
        return 0;
    }
}

static
crembx(char *name)
{
    /*
     * Create a mailbox and return its name.
     */
    short chan;
    int status;

    status = sys$crembx(0, &chan, 0, 0, 0, 0, 0, 0);
    if (status & 1) {
	if (getDeviceName(chan, name))
	    return chan;
	else {
	    delmbx(chan);
	    return 0;
	}
    }
    else {
	setVMSerrorCode(status);
	return 0;
    }
}

static
delmbx(int chan)
{
    int status;

    status = sys$delmbx(chan);
    if (status & 1)
	return 1;
    else {
	setVMSerrorCode(status);
	return 0;
    }
}

static
cleanmbx(int chan)
{
    /*
     * Read a message in our mailbox
     */
    char buff[512];
    short iosb[4];
    short brdcstCount;
    int status;
    struct {
	short msgType;
	short unitNumber;
	char count;
	char controller[15];
    } msg;

    status = sys$qiow(0, chan, IO$_READVBLK, iosb, 0, 0,
		      &msg, sizeof(msg), 0, 0, 0, 0);
    if ((status & 1) && ((status = iosb[0]) & 1)) {
	if (msg.msgType == MSG$_TRMBRDCST) {
	    /*
	     * Suck up broadcast message
	     */
	    status = sys$qiow(0, chan, IO$_READVBLK, iosb, 0, 0,
			      &brdcstCount, 2, 0, 0, 0, 0);
	    status = sys$qiow(0, chan, IO$_READVBLK, iosb, 0, 0,
			      buff, &brdcstCount, 0, 0, 0, 0);
	}
    }
    else
	printf("Could not clean mailbox: %x\n", status);
}

getDeviceType(int chan)
{
    int status;
    int retLength;
    struct {
	short l;
	short code;
	void *a;
	int *retLength;
    } dviItemList[2];
    union devdef deviceChar;

    dviItemList[0].l = sizeof(deviceChar);
    dviItemList[0].code = DVI$_DEVCHAR;
    dviItemList[0].a = &deviceChar;
    dviItemList[0].retLength = &retLength;
    dviItemList[1].l = dviItemList[1].code = 0;
    status = sys$getdviw(0, chan, 0, &dviItemList, 0, 0, 0, 0);
    if (status & 1) {
	if (deviceChar.dev$v_trm)
	    return DEV_IS_TERM;
	else if (deviceChar.dev$v_mbx)
	    return DEV_IS_MBX;
	else
	    return 0;
    }
    else {
	setVMSerrorCode(status);
	return -1;
    }
}

getDeviceTypeByName(char *devName)
{
    /*
     * Given a device name, return what kind it is
     */
    int status;
    int retLength;
    struct {
	short l;
	short code;
	void *a;
	int *retLength;
    } dviItemList[2];
    union devdef deviceChar;
    DESC dn;

    LOAD_DESC(dn, devName);
    dviItemList[0].l = sizeof(deviceChar);
    dviItemList[0].code = DVI$_DEVCHAR;
    dviItemList[0].a = &deviceChar;
    dviItemList[0].retLength = &retLength;
    dviItemList[1].l = dviItemList[1].code = 0;
    status = sys$getdviw(0, 0, &dn, &dviItemList, 0, 0, 0, 0);
    if (status & 1) {
	if (deviceChar.dev$v_trm)
	    return DEV_IS_TERM;
	else if (deviceChar.dev$v_mbx)
	    return DEV_IS_MBX;
	else
	    return 0;
    }
    else {
	setVMSerrorCode(status);
	return -1;
    }
}

static
getDeviceName(int chan, char *name)
{
    int status;
    int retLength = 0;
    struct {
	short l;
	short code;
	void *a;
	int *retLength;
    } dviItemList[2];

    dviItemList[0].l = 64;
    dviItemList[0].code = DVI$_DEVNAM;
    dviItemList[0].a = name;
    dviItemList[0].retLength = &retLength;
    dviItemList[1].l = dviItemList[1].code = 0;
    status = sys$getdviw(0, chan, 0, &dviItemList, 0, 0, 0, 0);
    if (status & 1) {
	name[retLength] = '\0';
	return  1;
    }
    else {
	setVMSerrorCode(status);
	return 0;
    }
}

static int
setWriteAttention(int chan, void (*devReadyAstProc)(ClientData), ClientData clientdata)
{
    short iosb[4];
    int status;

    status = sys$qiow(0, chan, IO$_SETMODE|IO$M_WRTATTN, iosb, 0, 0,
		      devReadyAstProc, clientdata, 0, 0, 0, 0);
    return 1;
}

static short outfd = -1;

static void
debugOut(char *msg)
{
    if (outfd == -1) {
	outfd = assignChan("sys$output", (char *)0);
    }
    sys$qiow(0, outfd, IO$_WRITEVBLK, 0, 0, 0,
			  "\n", 1, 0, 0, 0, 0);
    sys$qiow(0, outfd, IO$_WRITEVBLK, 0, 0, 0,
			  msg, strlen(msg), 0, 0, 0, 0);
    sys$qiow(0, outfd, IO$_WRITEVBLK, 0, 0, 0,
			  "\r\n", 2, 0, 0, 0, 0);
}

/*
 *----------------------------------------------------------------------
 *
 * TtySetOptionProc --
 *
 *	Sets an option on a channel.
 *
 * Results:
 *	A standard Tcl result. Also sets interp->result on error if
 *	interp is not NULL.
 *
 * Side effects:
 *	May modify an option on a device.
 *
 *----------------------------------------------------------------------
 */

static int		
TtySetOptionProc(instanceData, interp, optionName, value)
    ClientData instanceData;	/* File state. */
    Tcl_Interp *interp;		/* For error reporting - can be NULL. */
    char *optionName;		/* Which option to set? */
    char *value;		/* New value for option. */
{
    DevState *fsPtr = (DevState *) instanceData;
    unsigned int len;
    TtyAttrs tty;

    len = strlen(optionName);
    if ((len > 1) && (strncmp(optionName, "-mode", len) == 0)) {
	if (TtyParseMode(interp, value, &tty.baud, &tty.parity, &tty.data,
		&tty.stop) != TCL_OK) {
	    return TCL_ERROR;
	}
	/*
	 * system calls should be checked there. -- dl
	 */

	TtySetAttributes(fsPtr->chan, &tty);
	return TCL_OK;
    } else {
    	return Tcl_BadChannelOption(interp, optionName, "mode");
    }
}

/*
 *----------------------------------------------------------------------
 *
 * TtyGetOptionProc --
 *
 *	Gets a mode associated with an IO channel. If the optionName arg
 *	is non NULL, retrieves the value of that option. If the optionName
 *	arg is NULL, retrieves a list of alternating option names and
 *	values for the given channel.
 *
 * Results:
 *	A standard Tcl result. Also sets the supplied DString to the
 *	string value of the option(s) returned.
 *
 * Side effects:
 *	The string returned by this function is in static storage and
 *	may be reused at any time subsequent to the call.
 *
 *----------------------------------------------------------------------
 */

static int		
TtyGetOptionProc(instanceData, interp, optionName, dsPtr)
    ClientData instanceData;	/* File state. */
    Tcl_Interp *interp;		/* For error reporting - can be NULL. */
    char *optionName;		/* Option to get. */
    Tcl_DString *dsPtr;		/* Where to store value(s). */
{
    DevState *fsPtr = (DevState *) instanceData;
    unsigned int len;
    char buf[32];
    TtyAttrs tty;

    if (optionName == NULL) {
	Tcl_DStringAppendElement(dsPtr, "-mode");
	len = 0;
    } else {
	len = strlen(optionName);
    }
    if ((len == 0) || 
	    ((len > 1) && (strncmp(optionName, "-mode", len) == 0))) {
	TtyGetAttributes(fsPtr->chan, &tty);
	sprintf(buf, "%d,%c,%d,%d", tty.baud, tty.parity, tty.data, tty.stop);
	Tcl_DStringAppendElement(dsPtr, buf);
	return TCL_OK;
    } else {
    	return Tcl_BadChannelOption(interp, optionName, "mode");
    }
}


#include <ttdef.h>
#include <tt2def.h>

static struct {int baud; unsigned long speed;} speeds[] = {
    {50, TT$C_BAUD_50},
    {75, TT$C_BAUD_75},
    {110, TT$C_BAUD_110},
    {134, TT$C_BAUD_134},
    {150, TT$C_BAUD_150},
    {300, TT$C_BAUD_300},
    {600, TT$C_BAUD_600},
    {1200, TT$C_BAUD_1200},
    {1800, TT$C_BAUD_1800},
    {2000, TT$C_BAUD_2000},
    {2400, TT$C_BAUD_2400},
    {3600, TT$C_BAUD_3600},
    {4800, TT$C_BAUD_4800},
    {7200, TT$C_BAUD_7200},
    {9600, TT$C_BAUD_9600},
    {19200, TT$C_BAUD_19200},
    {38400, TT$C_BAUD_38400},
    {-1, 0}
};

/*
 *---------------------------------------------------------------------------
 *
 * TtyGetSpeed --
 *
 *	Given a baud rate, get the mask value that should be stored in
 *	the termios, termio, or sgttyb structure in order to select that
 *	baud rate.
 *
 * Results:
 *	As above.
 *
 * Side effects:
 *	None.
 *
 *---------------------------------------------------------------------------
 */

static unsigned long
TtyGetSpeed(baud)
    int baud;			/* The baud rate to look up. */
{
    int bestIdx, bestDiff, i, diff;
    
    bestIdx = 0;
    bestDiff = 1000000;
    
    /*
     * If the baud rate does not correspond to one of the known mask values,
     * choose the mask value whose baud rate is closest to the specified
     * baud rate.
     */

    for (i = 0; speeds[i].baud >= 0; i++) {
	diff = speeds[i].baud - baud;
	if (diff < 0) {
	    diff = -diff;
	}
	if (diff < bestDiff) {
	    bestIdx = i;
	    bestDiff = diff;
	}
    }
    return speeds[bestIdx].speed;
}

/*
 *---------------------------------------------------------------------------
 *
 * TtyGetBaud --
 *
 *	Given a speed mask value from a termios, termio, or sgttyb
 *	structure, get the baus rate that corresponds to that mask value.
 *
 * Results:
 *	As above.  If the mask value was not recognized, 0 is returned.
 *
 * Side effects:
 *	None.
 *
 *---------------------------------------------------------------------------
 */

static int
TtyGetBaud(speed)
    unsigned long speed;	/* Speed mask value to look up. */
{
    int i;
    
    for (i = 0; speeds[i].baud >= 0; i++) {
	if (speeds[i].speed == speed) {
	    return speeds[i].baud;
	}
    }
    return 0;
}


/*
 *---------------------------------------------------------------------------
 *
 * TtyInit --
 *
 *	Given file descriptor that refers to a serial port, 
 *	initialize the serial port to a set of sane values so that
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Serial device initialized.
 *
 *---------------------------------------------------------------------------
 */
static void
TtyInit(fd)
    int fd;			/* Open file descriptor for serial port to
				 * be initialized. */
{
#ifdef USE_TERMIOS
    struct termios termios;

    tcgetattr(fd, &termios);
    termios.c_iflag = IGNBRK;
    termios.c_oflag = 0;
    termios.c_lflag = 0;
    termios.c_cflag |= CREAD;
    termios.c_cc[VMIN] = 60;
    termios.c_cc[VTIME] = 2;
    tcsetattr(fd, TCSANOW, &termios);
#else	/* !USE_TERMIOS */
#ifdef USE_TERMIO
    struct termio termio;

    ioctl(fd, TCGETA, &termio);
    termio.c_iflag = IGNBRK;
    termio.c_oflag = 0;
    termio.c_lflag = 0;
    termio.c_cflag |= CREAD;
    termio.c_cc[VMIN] = 60;
    termio.c_cc[VTIME] = 2;
    ioctl(fd, TCSETAW, &termio);
#else	/* !USE_TERMIO */
#ifdef USE_SGTTY
    struct sgttyb sgttyb;

    ioctl(fd, TIOCGETP, &sgttyb);
    sgttyb.sg_flags &= (EVENP | ODDP);
    sgttyb.sg_flags |= RAW;
    ioctl(fd, TIOCSETP, &sgttyb);
#endif	/* USE_SGTTY */
#endif	/* !USE_TERMIO */
#endif	/* !USE_TERMIOS */
}

/*
 *---------------------------------------------------------------------------
 *
 * TtyGetAttributes --
 *
 *	Get the current attributes of the specified serial device.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *---------------------------------------------------------------------------
 */
 
static void
TtyGetAttributes(fd, ttyPtr)
    int fd;			/* Open file descriptor for serial port to
				 * be queried. */
    TtyAttrs *ttyPtr;		/* Buffer filled with serial port
				 * attributes. */
{
    unsigned char c;
    short iosb[4];
    int status;
    int charBuff[3];

    status = sys$qiow(0, fd, IO$_SENSEMODE, &iosb, 0, 0,
			  charBuff, 12, 0, 0, 0, 0);
    if (status & 1) {
    	ttyPtr->baud = TtyGetBaud(iosb[1] & 0xffff);
    	c = iosb[3] & 0xffff;
    	if (c & TT$M_PARITY) {
	    if (c & TT$M_ODD)
		ttyPtr->parity = 'o';
	    else
	    	ttyPtr->parity = 'e';
	}
	else
	    ttyPtr->parity = 'n';
	/* don't know about mark, set */
	ttyPtr->data = 8;
	ttyPtr->stop = 1;
    }
    else {
	ttyPtr->baud = 0;
	ttyPtr->parity = 'n';
	ttyPtr->data = 0;
	ttyPtr->stop = 0;
    }
}

/*
 *---------------------------------------------------------------------------
 *
 * TtySetAttributes --
 *
 *	Set the current attributes of the specified serial device. 
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *---------------------------------------------------------------------------
 */
 
static void
TtySetAttributes(fd, ttyPtr)
    int fd;			/* Open file descriptor for serial port to
				 * be modified. */
    TtyAttrs *ttyPtr;		/* Buffer containing new attributes for
				 * serial port. */
{
    short iosb[4];
    int status;
    int charBuff[3];
    int parity;
    int mods;

    if (ttyPtr->parity != 'n') {
	status = sys$qiow(0, fd, IO$_SENSEMODE, &iosb, 0, 0,
			  charBuff, 8, 0, 0, 0, 0);

	/*
	 * Is parity already set?
	 */
	parity = TT$M_ALTRPAR;
	if (status & 1) {
	    if (!(iosb[3] & TT$M_PARITY)) {
	    	parity |= TT$M_PARITY;
	    }
    	}
    	if (ttyPtr->parity == 'o')
    	    parity |= TT$M_ODD;
    	else if (ttyPtr->parity == 'e')
    	    ;
    }
    else
    	parity = 0;
    status = sys$qiow(0, fd, IO$_SETMODE, &iosb, 0, 0,
			  (void *)0, 0, ttyPtr->baud, 0, parity, 0);
}

/*
 *---------------------------------------------------------------------------
 *
 * TtyParseMode --
 *
 *	Parse the "-mode" argument to the fconfigure command.  The argument
 *	is of the form baud,parity,data,stop.
 *
 * Results:
 *	The return value is TCL_OK if the argument was successfully
 *	parsed, TCL_ERROR otherwise.  If TCL_ERROR is returned, an
 *	error message is left in interp->result (if interp is non-NULL).
 *
 * Side effects:
 *	None.
 *
 *---------------------------------------------------------------------------
 */
 
static int
TtyParseMode(interp, mode, speedPtr, parityPtr, dataPtr, stopPtr)
    Tcl_Interp *interp;		/* If non-NULL, interp for error return. */
    CONST char *mode;		/* Mode string to be parsed. */
    int *speedPtr;		/* Filled with baud rate from mode string. */
    int *parityPtr;		/* Filled with parity from mode string. */
    int *dataPtr;		/* Filled with data bits from mode string. */
    int *stopPtr;		/* Filled with stop bits from mode string. */
{
    int i, end;
    char parity;
    static char *bad = "bad value for -mode";

    i = sscanf(mode, "%d,%c,%d,%d%n", speedPtr, &parity, dataPtr,
	    stopPtr, &end);
    if ((i != 4) || (mode[end] != '\0')) {
	if (interp != NULL) {
	    Tcl_AppendResult(interp, bad, ": should be baud,parity,data,stop",
		    NULL);
	}
	return TCL_ERROR;
    }
    if (strchr("noems", parity) == NULL) {
	if (interp != NULL) {
	    Tcl_AppendResult(interp, bad,
		    " parity: should be n, o, e, m, or s", NULL);
	}
	return TCL_ERROR;
    }
    *parityPtr = parity;
    if ((*dataPtr < 5) || (*dataPtr > 8)) {
	if (interp != NULL) {
	    Tcl_AppendResult(interp, bad, " data: should be 5, 6, 7, or 8",
		    NULL);
	}
	return TCL_ERROR;
    }
    if ((*stopPtr < 0) || (*stopPtr > 2)) {
	if (interp != NULL) {
	    Tcl_AppendResult(interp, bad, " stop: should be 1 or 2", NULL);
	}
	return TCL_ERROR;
    }
    return TCL_OK;
}
